// ESFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.space.Grid2dImpl;
import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.collections.ArrayImpl;
import swarm.collections.ListShufflerImpl;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class ESFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the total number of operating Units we are using*/
    public int potentialUnitTypes;
    /** the probability of initial generation of a unit, listed into the 
     *  unit directory */
    public double unitGenerationInitialP;
    /** repetition in unit creation 
     *  if potentialUnitNumberPerType > 1 each type of unit (as in each row of 
     *  unitData/unitBasicData.txt) can be be created in multiple copies, 
     *  always applying unitGenerationInitialP; unit number are 
     *  unit*100+repetition factor (if  potentialUnitNumberPerType > 99 we have 
     *  nonsense numbers)*/
    public int potentialUnitNumberPerType;
    /* the max number of steps to be done to complete an order */
    /** probability of unit new Unit generation while the simulation is running */
    public double newUnitGenerationP;

    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time) */
    public int maxStepLength;
    /** the total production time used by done orders */
    public float totalProductionTime;
    /** the total length of the recipes of the done orders */
    public float totalRecipeLength;
    /** the X size  of the unit space */
    public int unitSpaceXSize;
    /** the Y size  of the unit space */
    public int unitSpaceYSize;
    /** the unit space (a metaphorical one) */
    public Grid2dImpl unitSpace;
    /** the unit activity space, always metaphoric */
    public UnitActivitySpace unitActivitySpace;
    /** the min starting visibility */
    public double minStartingVisibility;
    /** the max starting visibility */
    public double maxStartingVisibility;
    /** the min level of inter-visibility; below this level two units cannot
     *  exchange products */
    public double interVisibilityMinLevel;
    /** increasing visibility in each step 1 of each unit */
    public double increasingVisibilityStep;
    /** max inactivity value */
    public int maxInactivity;
    /** max quantity of unsent products */
    public int maxUnsentProducts;
    /** potential production from the beginning of the run */
    public float potentialProduction;
    /** global production from the beginning of the run */
    public float globalProduction;
    /** concluded orders form the beginning of the simulation */
    public float concludedOrders;
    /** create the display of the numberOfUnits */
    public boolean createNumberOfUnitsGraph;
    /** create the display of the production */
    public boolean createProductionGraph;
    /** create the display of the ratios */
    public boolean createRatioGraph;

    /** the layer of this model */
    public int layer;

    /** the order generator */
    public OrderGenerator orderGenerator;

    /** 
     * ActionGroup for holding an ordered sequence of action 
     * (the numbers of the model action names are the same used in the full
     * jES model, so some are lacking in this version)
    */
    public ActionGroup modelActions2, modelActions2b, 
	modelActions2generator;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** its iterator */
    public ListIndex unitListIndex;

    /** a shuffler for the list */
    public ListShufflerImpl myShuffler;

    /** the list of the pro forma units (all the "would be" units) */
    public ListImpl proFormaUnitList;

    /** the tool to assign order to units */
    AssigningTool assigningTool;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameModelSwarm$1$ESFrameModelProbeMap)
     */
    Object modelProbe;

    // the constructor of the model
    public ESFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	potentialUnitTypes = 10;
	unitGenerationInitialP=0.9;
        potentialUnitNumberPerType=1;
	newUnitGenerationP=0.0;
	maxStepNumber=30;
	maxStepLength=1;
	unitSpaceXSize=100;
	unitSpaceYSize=100;
	minStartingVisibility=0.0;
	maxStartingVisibility=100.0;
	interVisibilityMinLevel=0.0;
	increasingVisibilityStep=0.0;
	maxInactivity=10;
        maxUnsentProducts=10;
	globalProduction=0;
	potentialProduction=0;
	concludedOrders=0;
	createNumberOfUnitsGraph=true;
	createProductionGraph=true;
	createRatioGraph=true;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameModelSwarm$1$ESFrameModelProbeMap.class
	class ESFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public ESFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("potentialUnitTypes");
		addVar ("unitGenerationInitialP");
		addVar ("potentialUnitNumberPerType");
		addVar ("newUnitGenerationP");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
		addVar ("unitSpaceXSize");
		addVar ("unitSpaceYSize");
		addVar ("minStartingVisibility");
		addVar ("maxStartingVisibility");
		addVar ("interVisibilityMinLevel");
		addVar ("increasingVisibilityStep");
		addVar ("maxInactivity");
		addVar ("maxUnsentProducts");
		addVar ("createNumberOfUnitsGraph");
		addVar ("createProductionGraph");
		addVar ("createRatioGraph");
    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	modelProbe=new ESFrameModelProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameModelProbeMap) modelProbe, getClass ());
    }
    /** Build the model objects. */
    public Object buildObjects ()
    {
	int i, rep;

	// allow our parent class to build anything
	super.buildObjects();

	// check visibility
	if(maxStartingVisibility<minStartingVisibility)
	    {
		System.out.println(
				   "maxStartingVisibility < minStartingVisibility");
		MyExit.exit(1);
	    }
			

        // create the class to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), potentialUnitTypes, layer);
	unitParameters.readUnitData();

	// the list of the production units
	unitList = new ListImpl (getZone());
	// its iterator
	unitListIndex = unitList.listBegin(getZone());

	// the shuffler
	myShuffler= new ListShufflerImpl(getZone());

	// the list of the pro forma units
	proFormaUnitList = new ListImpl (getZone());

	// create the unit activity space (which is automatically filled with 0 */
	unitActivitySpace = new UnitActivitySpace(Globals.env.globalZone, 
					      unitSpaceXSize, unitSpaceYSize);

	// create a 2-D grid that will be used to keep track of
	// each unit metaphorical position
	unitSpace = new Grid2dImpl(Globals.env.globalZone, 
				   unitSpaceXSize, unitSpaceYSize);
	unitSpace.fastFillWithObject(null);

	// the instance of assigningTool
	assigningTool = new AssigningTool (getZone(), unitList, unitListIndex,
					   unitActivitySpace,
					   interVisibilityMinLevel);

	// the orderGenerator (with pro forma unit list to build the full
	// dictionary
	orderGenerator = new OrderGenerator
	    (getZone(), maxStepNumber, maxStepLength, proFormaUnitList,
	     this, assigningTool);


	// create the pro forma units (used to build the activity dictionary (containing 
	// the steps to be included in the recipes) in OrderGenerator
	for (i = 1; i <= potentialUnitTypes; i++)
	    {
		Unit aProFormaUnit;
		
		aProFormaUnit = new Unit  (getZone(), this, unitActivitySpace, 
					   assigningTool, (int) 0, (int) 0, (byte) 0, 
					   (int) 0, (int) 0);
		unitParameters.setParametersTo(i, 0, aProFormaUnit);
		
		// add the unit to the list
		proFormaUnitList.addLast (aProFormaUnit);
	    }

	if (potentialUnitNumberPerType == 0) potentialUnitNumberPerType=1;
	// create the units
	for (i = 1; i <= potentialUnitTypes; i++)
	    for (rep=1; rep <= potentialUnitNumberPerType; rep++)
		if(unitGenerationInitialP >= 
		   Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
		    generateUnit(i, rep);
  	
	if (unitList.getCount()==0)
	    {
		System.out.println("General error, no unit generated");
		MyExit.exit(1);
	    }

	// creating the dictionary in orderGenerator
	orderGenerator.setDictionary();

	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i, ii;
	super.buildActions();

	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions2generator = new ActionGroupImpl (getZone ());
	modelActions2b = new ActionGroupImpl (getZone ());

	/* cleaning the unit activity space */
	modelActions2.createActionTo$message(unitActivitySpace,
			         SwarmUtils.getSelector(unitActivitySpace,"fillWithZero"));

	/* producing */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	modelActions2generator.createActionTo$message(orderGenerator,
			       SwarmUtils.getSelector(orderGenerator,
				     "createRandomOrderWithNSteps"));
	modelActions2b.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep2"));

	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"dropUnits"));
	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"generateANewUnit"));
	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"shuffleUnitList"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), 1);

	modelSchedule.at$createAction (0, modelActions2);
	modelSchedule.at$createAction (0, modelActions2generator);
	modelSchedule.at$createAction (0, modelActions2b);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** unit generation */
    public void generateUnit(int i, int rep)
    {
	int xPos, yPos, xD=0, yD=0, count;
	Unit aUnit;
	byte aColor = (byte) 1;
	
	xPos=Globals.env.uniformIntRand.getIntegerWithMin$withMax(
								  0, unitSpaceXSize-1);
	yPos=Globals.env.uniformIntRand.getIntegerWithMin$withMax(
								  0, unitSpaceYSize-1);
	
	// if the destination target is not empty, chose a direction ...
	if (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		xD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		yD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		
		while (xD==0 && yD==0)
		    {		
			xD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
			yD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
		    }
	    }
	// ... and move
	count=0;
	while (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		count++;
		// unitSpaceXSize+unitSpaceYSize is a raw test
		if (count >= unitSpaceXSize + unitSpaceYSize)
		    {
			System.out.println("Unable to find a place for unit "+
					   i + " after " + count + 
					   " displacements");
			MyExit.exit(1);
		    }
		xPos += xD; xPos=CoordinateManager.check(unitSpaceXSize, xPos);	    
		yPos += yD; yPos=CoordinateManager.check(unitSpaceYSize, yPos);
	    }
	
	aUnit = new Unit  (getZone(), this, unitActivitySpace, 
			   assigningTool, xPos, yPos, aColor,
			   Globals.env.uniformDblRand.
			   getDoubleWithMin$withMax(
						     minStartingVisibility, 
						     maxStartingVisibility),
			   increasingVisibilityStep);
	unitParameters.setParametersTo(i, rep, aUnit);
	
	// add the unit to the list
	unitList.addLast (aUnit);
	// placing the unit into the space
	unitSpace.putObject$atX$Y(aUnit, xPos, yPos);
	
    } // end generateUnit

    public void generateANewUnit(){
	int r, i;

	if(newUnitGenerationP >= 
		   Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		r=Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(0,proFormaUnitList.getCount() -1);

		i=((Unit)proFormaUnitList.atOffset(r)).getProductionPhase();

		generateUnit(i, 0); // no rep value, so the unit number will be i*100
	    }
    }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** shuffling the unitList */
    public void shuffleUnitList(){
	myShuffler.shuffleWholeList(unitList);

	/*
	// to check the effect of the shuffling action
	  unitListIndex.setOffset(0);
	  for (int i=0;i<unitList.getCount();i++)
	    {
		System.out.print(((Unit)unitListIndex.get()).getUnitNumber()+" ");
		unitListIndex.next();
	    }
	System.out.println(" ");
	*/
    }

    /** dropping units */
    public void dropUnits()
    {

	int i, nu, j, unsent=0, waiting=0, unsentValue=0, waitingValue=0;
	Unit aUnit;
	Order anOrder;

	nu=unitList.getCount();

	for (i=0;i<nu;i++)
	    {
		// the || is introduced to take in consideration a rare but possible
		// situation: the inactvity is at a limit value, but an order is just arrived
		aUnit=(Unit) unitList.removeFirst();
		if((aUnit.getInactivity() <= maxInactivity 
		    || aUnit.getWaitingListLength()>0) &&
		   aUnit.getUnsentProducts() <= maxUnsentProducts)unitList.addLast(aUnit);
		else
		    {
			unitSpace.putObject$atX$Y(null, aUnit.getXPos(), aUnit.getYPos());

			unsentValue=0;
			unsent=aUnit.getUnsentProducts();
			for (j=1;j<=unsent;j++)
			    {
			    anOrder = (((Order)((ListImpl)aUnit.getDailyProductionList()).removeFirst()));
			    unsentValue += anOrder.getNumberOfDirectStepsDone();
			    anOrder.drop();
			    if(StartESFrame.verbose)
				System.out.println(anOrder.getOrderNumber()+" u  "+unsentValue);
			    }

			waitingValue=0;
			waiting=aUnit.getWaitingListLength();
			for (j=1;j<=waiting;j++)
			    {
			    anOrder = (((Order)((ListImpl)aUnit.getWaitingList()).removeFirst()));
			    waitingValue += anOrder.getNumberOfDirectStepsDone();
			    anOrder.drop();
			    if(StartESFrame.verbose)
				System.out.println(anOrder.getOrderNumber()+" w  "+waitingValue);
			    }

			addToGlobalProduction(-unsentValue-waitingValue);
			aUnit.drop();
		    }
	    }

	if(unitList.getCount()==0)
	    {
		System.out.println();
		System.out.println("Last unit dropped in layer "+layer+"; press Quit to exit.");
		StartESFrame.eSFrameObserverSwarm.
		    getControlPanel().setStateStopped();
	    }


    }


    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(float pt, float rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartESFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return totalProductionTime/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

    /** report the number of units */
    public float getNumberOfUnits(){
	return (float) unitList.getCount();
    }

    /** return the unitSpace */
    public Grid2dImpl getUnitSpace()
    {
	return unitSpace;
    }

    /** increasing potential production value */
    public void addToPotentialProduction(int d){
	 potentialProduction += d;
    }

    /** reporting potential production value */
    public float getPotentialProduction(){
	return potentialProduction;
    }

    /** increasing/decreasing global production value */
    public void addToGlobalProduction(int d){
	globalProduction += d;
    }

    /** reporting the global production value */
    public float getGlobalProduction(){
	return globalProduction;
    }

    /** increasing/decreasing the value of the concluded orders */
    public void addToConcludedOrders(float d){
	concludedOrders += d;
    }

    /** reporting the value of the concluded orders */
    public float getConcludedOrderSteps(){
	return concludedOrders;
    }
    /** return the unitActivitySpace address */
    public UnitActivitySpace getUnitActivitySpace()
    {
	return unitActivitySpace;
    }

    /** setting the layer of this model */
    public void setLayer(int l)
    {
	layer=l;
    }

    /** return if we want to create the numberOfUnitsGraph */
    public boolean getCreateNumberOfUnitsGraph()
    {
	return createNumberOfUnitsGraph;
    }

    /** return if we want to create the productionGraph */
    public boolean getCreateProductionGraph()
    {
	return createProductionGraph;
    }

    /** return if we want to create the ratioGraph */
    public boolean getCreateRatioGraph()
    {
	return createRatioGraph;
    }

}


